/* smta - Copyright (C) 2003-4 bakunin - Andrea Marchesini 
 *                             <bakunin@autistici.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#else
# error Use configure; make; make install
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <termios.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <string.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <errno.h>

#ifdef ENABLE_GLIB
#include <glib.h>
#endif

#define MAX_PASSWORD 80
#define LISTEN 2
#define AUTHOR "Andrea Marchesini - bakunin@autistici.org"
#define COPYRIGHT "CopyRight (c) 2004 - GPL"

#define MSMTP 1
#define SSMTP 2

#define get_random(min, max) ((rand() % (int)(((max)+1) - (min))) + (min))

struct termios tin_n, tin_o;
struct termios tout_n, tout_o;
struct termios terr_n, terr_o;

/* Prototipes */
int usage(int);
char *trim(char *);

void echo_off(void);
void echo_on(void);

char *tempfile(void);
int get_password(int, char *, size_t);
void set_exit(int);
void scheduler(int, char *, int);

char *socket_default(void);
int socket_client(char *);
int socket_daemon(char *, int);
void socket_close(int);

int mta_check(char *);
char *mta_get_line(FILE *);
char *mta_msmtp(char *, int, int, char **);
int mta_exec(char **);
int mta_run(int , char *, int, int, char **);
void mta_remove(char *);

/* Usage */
int usage(int ret) {
  fprintf(stderr, "%s %s\n\n", PACKAGE, VERSION);
  fprintf(stderr, "\t%s [-d [-l <numer>] [-t <timer>] [-b]] | [[-e]|[-mta <mta>]] [-sock]\n\n", PACKAGE);
  fprintf(stderr, "-d | --daemon       daemon function\n");
  fprintf(stderr, "-l | --listen       listen number [default: %d]\n", LISTEN);
  fprintf(stderr, "-t | --timer        timer [default: 0=disactivated]\n");
  fprintf(stderr, "-b | --background   background\n");
  fprintf(stderr, "-e | --exit         close a daemon\n");
  fprintf(stderr, "-s | --socket       unix socket [default %s]\n", socket_default());
  fprintf(stderr, "-m | --mta          msmtp|ssmtp\n"); 
  fprintf(stderr, "-h | --help         this help\n");
  fprintf(stderr, "\n%s %s\n\n",AUTHOR, COPYRIGHT);
  return ret;
}

/* Realy simple trim function */
char *
trim (char *tmp)
{
  int i = 0;
  while (tmp[i] == ' ' || tmp[i] == '\t' || tmp[i] == '\r' || tmp[i] == '\n')
    tmp++;

  i = strlen (tmp);
  i--;

  while (tmp[i] == ' ' || tmp[i] == '\t' || tmp[i] == '\r' || tmp[i] == '\n')
    i--;

  tmp[i + 1] = 0;

  return tmp;
}

/* for get the password, disactive/active the echo */
void
echo_off (void)
{
  tcgetattr (STDIN_FILENO, &tin_o);
  tin_n = tin_o;

  tin_n.c_lflag &= ~(ECHO);
  tcsetattr (STDIN_FILENO, TCSADRAIN, &tin_n);
}

void
echo_on (void)
{
  tcsetattr (STDIN_FILENO, TCSADRAIN, &tin_o);
}

/* Search the tempfile... */
char *tempfile(void) {
  char *tmp;
  int d=0;
  struct stat st;
  char buf[4];

#ifdef ENABLE_GLIB
  if(!(tmp=(char *)malloc((strlen((char *)g_get_tmp_dir())+8)*sizeof(char)))) return NULL;
  sprintf(tmp, "%s/.smta",(char *)g_get_tmp_dir());
#else
  if(tmp=(char *)malloc(14*sizeof(char))) return NULL;
  sprintf(tmp, "/tmp/.smta");
#endif	

  /* check the file /tmp/.smtaNUM */
  while(d<1000) {
    sprintf(buf, "%d",d);
    strcat(tmp, buf);

    if(lstat(tmp, &st)) break;

    d++;
  }

  if(d==1000) {
    free(tmp);
    return NULL;
  }

  return tmp;
}

/* Get password function */
int get_password(int fd, char *password, size_t size) {
  int ret;
  
  /* This is a realy simple protocol:
   * client: PASSWORD
   * server: write the password
   * 
   * or
   *
   * client: EXIT
   * server: close it self */

  if(write(fd, "PASSWORD",8)<0) {
    fprintf(stderr, "protocol error\n");
    return 1;
  }

  if((ret=read(fd, password, size))<0) {
    fprintf(stderr, "protocol error\n");
    return 1;
  }

  if(password[ret-1]=='\n') ret--;
  password[ret]=0;

  return 0;
}

/* Close the connection */
void set_exit(int fd) {
  
  if(write(fd, "EXIT",4)<0)
    fprintf(stderr, "protocol error\n");
}

/* The daemon scheduler */
void scheduler(int fd, char *password, int timer) {
  int c;
  struct sockaddr_un incoming;
  size_t size_un = sizeof (struct sockaddr_un);
  char buf[200];
  int ret;
  int quit=0;
  struct timeval start, now;

  /* If there is the timer variabile... */
  if(timer)
    gettimeofday (&start, NULL);

  while(!quit) {
    usleep(10000);

    gettimeofday (&now, NULL);

    /* Check the timer, and close */
    if(timer && (((double)now.tv_sec - (double) start.tv_sec) + ((double) now.tv_usec - (double) start.tv_usec) / 1000000.0)>=(double)timer) 
      break;

    if ((c = accept (fd, (struct sockaddr *) &incoming, &size_un)) < 0 &&
         errno==EAGAIN)
      continue;
    
    if((ret=read(c,buf, sizeof(buf)))<0) {
      close(c);
      continue;
    }

    /* Exit */
    if(!strncmp(buf,"EXIT",ret)) quit=1;

    /* Write the password */
    else if(!strncmp(buf,"PASSWORD",ret))
      write(c, password, strlen(password));

    close(c);
  }
}

/* Default socket for smta: ~/.smta.sock */
char *socket_default(void) {
  char *home;
  char *buf;

  if(!(home=getenv("HOME"))) return "/tmp/smta.sock";

  if(!(buf=(char *)malloc((strlen(home)+11)*sizeof(char))))
     return "/tmp/smta.sock";

  sprintf(buf, "%s/.smta.sock", home);

  return buf;
}

/* client -> daemon */
int socket_client(char *sock) {
  struct sockaddr_un saddr;
  int fd;

  if((fd = socket (AF_UNIX, SOCK_STREAM, 0)) < 0) {
    fprintf(stderr, "socket error\n");
    return -1;
  }

  saddr.sun_family = AF_UNIX;
  strncpy (saddr.sun_path, sock, 108);

  if(connect (fd, (struct sockaddr *) &saddr, sizeof (saddr)) == -1) {
    fprintf(stderr, "connect to %s error\n", sock);
    close (fd);
    return -1;
  }

  return fd;
}

/* daemon -> others */
int socket_daemon(char *sock, int l) {
  int fd;
  struct sockaddr_un saddr;
  struct stat st;

  if(!lstat(sock, &st)) {
    unlink(sock);
  }

  if ((fd = socket (AF_UNIX, SOCK_STREAM, 0)) < 0) {
    fprintf(stderr, "socket error\n");
    return -1;
  }

  if (fcntl (fd, F_SETFL, O_NONBLOCK) < 0) {
    fprintf(stderr, "socket error\n");
    close(fd);
    return -1;
  }

  saddr.sun_family = AF_UNIX;
  strncpy (saddr.sun_path, sock, 108);
  
  if (bind (fd, (struct sockaddr *) &saddr, sizeof (saddr)) < 0) {
    fprintf(stderr, "binding of %s error\n",sock);
    close(fd);
    return -1;
  }

  if (listen (fd, l) < 0) {
    fprintf(stderr, "listen to %d error\n",l);
    close(fd);
    return -1;
  }

  return fd;
}

/* Close :) */
void socket_close(int fd) {
  close(fd);
}

/* Chec the mta parameter */
int mta_check(char *mta) {

  /* Now only this 2 mta: */
  if(!strcmp(mta, "msmtp")) return MSMTP;
  else if(!strcmp(mta, "ssmtp")) return SSMTP;

  return 0;
}

/* Get a single line (for search account information) */
char *mta_get_line(FILE *f) {
  static char buf[1024];
  char ch;
  int i=0;

  while((ch=getc(f))!=EOF) {

     if(ch=='\n') break;
     
     else buf[i++]=ch;
  }

  if(ch==EOF && !i) return NULL;

  buf[i]=0;

  return trim(buf);
}

/* Oky... complex :) */
char *mta_msmtp(char *password, int start, int stop, char **argv) {
  int k;
  char *account=NULL;
  char *file=NULL;
  char *tmp_file;
  FILE *fs, *fd;

  /* Search account and config file */
  for(k=start;k<stop;k++) {
     if(!strcmp(argv[k],"-a") && k<stop-1)
        account=argv[++k];

     else if(!strncmp(argv[k],"--account=",10))
        account=argv[k]+10;

     else if(!strcmp(argv[k],"--F") && k<stop-1)
        file=argv[++k];

     else if(!strncmp(argv[k],"--file=",7))
        file=argv[k]+7;
  }

  /* default config file (man msmtp: ~/.msmtprc) */
  if(!file) {
    char *home;

    if(!(home=getenv("HOME"))) {
      fprintf(stderr, "I can't read the HOME enviroment\n");
      return NULL;
    }

    if(!(file=(char *)malloc(sizeof(char)*(strlen(home)+10)))) {
      fprintf(stderr, "Error memory\n");
      return NULL;
    }

    sprintf(file, "%s/.msmtprc",home);
  }

  /* Get a tempfile */
  if(!(tmp_file=tempfile())) {
    fprintf(stderr, "Temp file: naahhh...\n");
    return NULL;
  }

  if(!(fs=fopen(file,"r"))) {
    fprintf(stderr, "I can't read the msmtp file %s.\n", file);
    return NULL;
  }

  if(!(fd=fopen(tmp_file,"w"))) {
    fprintf(stderr, "I can't read the tmp file %s.\n", tmp_file);
    return NULL;
  }

  /* If don't exist account element... only insert the password */
  if(!account) {
    char ch;

    while((ch=getc(fs))!=EOF) {
      putc(ch, fd);
    }

    fprintf(fd, "\npassword %s\n",password);
    fclose(fd);

    return tmp_file;
  }

  /* If exist account: */
  else {
    char *line;
    int k=0;

    while((line=mta_get_line(fs))) {

       if(!strncmp(line, "account ",8)) {
	 if(!k && !strcmp(account, line+8))  
           k++;
         else if(k) 
           break;
       } 

       if(k) {
         fputs(line,fd);
         fputc('\n',fd);
       }
    }

    fclose(fs);
    
    fprintf(fd, "\npassword %s\n",password);
    fclose(fd);

    return tmp_file;
  }
}

/* Exec the mta */
int mta_exec(char **a) {
  int ret=0;

  switch(fork()) {
    case -1: return 1;
    case 0:
      execvp (a[0], a);
      exit (1);
    default:
      wait (&ret);
  }

  return ret;
}

/* Run the mta */
int mta_run(int mta, char *pass, int start, int stop, char **argv) {
  int k, i=(stop-start)+1;
	  
  /* MSMTP: */
  if(mta==MSMTP) {
    char *array[i+2];
    char tmp[1024];
    char *tmp_file=mta_msmtp(pass, start, stop, argv);

    if(!tmp_file) return 1;

    snprintf(tmp,sizeof(tmp),"--file=%s",tmp_file);

    array[0]=MSMTP_CMD;
    array[1]=tmp;

    for(k=0;k<i;k++)
      array[k+2]=argv[start++];

    array[k+2]=NULL;

    k=mta_exec(array);
    mta_remove(tmp_file);

    return k;
  } 
  
  /* SSMTP */
  else if(mta==SSMTP) { 
    char *array[i+3];

    array[0]=SSMTP_CMD;
    array[1]="-ap";
    array[2]=pass;

    for(k=0;k<i;k++)
      array[k+3]=argv[start++];

    array[k+3]=NULL;

    return mta_exec(array);
  }

  /* next generation */
  else return 1;
}

/* remove the file */
void mta_remove(char *f) {
  struct stat st;
  int k;
  FILE *fl;

  srand(getpid());

  if(lstat(f, &st)) return;

  if(!(fl=fopen(f,"w"))) return;

  for(k=0;k<st.st_size;k++)
    putc(get_random(0,0xff),fl);

  fflush(fl);
  fclose(fl);

  unlink(f);
  
}

/* 3, 2, 1, go! */
int main(int argc, char *argv[]) {
  int ret;
  char password[MAX_PASSWORD];
  int l=0;
  char *sock=NULL;
  int daemon=0;
  int timer=0;
  int background=0;
  int e=0;
  int mta=0;
  int stop=0;
 
  for(ret=1;ret<argc;ret++) {
    if(!strcmp(argv[ret],"-t") || !strcmp(argv[ret],"--timer")) {

      if(++ret<argc)
        timer=atoi(argv[ret]);

      else return usage(1);
    }

    else if(!strcmp(argv[ret],"-l") || !strcmp(argv[ret],"--listen")) {

      if(++ret<argc)
        l=atoi(argv[ret]);

      else return usage(1);
    }

    else if(!strcmp(argv[ret],"-s") || !strcmp(argv[ret],"--socket")) {

      if(++ret<argc)
        sock=argv[ret];

      else return usage(1);
    }

    else if(!strcmp(argv[ret],"-e") || !strcmp(argv[ret],"--exit"))
      e=1;

    else if(!strcmp(argv[ret],"-d") || !strcmp(argv[ret],"--daemon"))
      daemon=1;

    else if(!strcmp(argv[ret],"-b") || !strcmp(argv[ret],"--background"))
      background=1;

    else if(!strcmp(argv[ret],"-m") || !strcmp(argv[ret],"--mta")) {

      if(++ret<argc) {
	     if(!(mta=mta_check(argv[ret])))
               return usage(1);

      } else return usage(1);
    }

    else if(!strcmp(argv[ret],"-h") || !strcmp(argv[ret],"--help"))
      return usage(0);

    else {
      stop=ret;
      break;
    }
  }

  /* Check argoments */
  if(daemon) {
     if(e || mta) return usage(1);
  } else {
    if((timer || background || l) || (!e && !mta) || (mta && !stop)) return usage(1);
  }
  
  /* Set the default sock */
  if(!sock) sock=socket_default();

  /* DAEMON! Open the connection */
  if(daemon) {
    if(!l) l=LISTEN;

    echo_off();
    fprintf(stderr, "Insert password for your smtp provider: ");

    ret=read(STDIN_FILENO, password, sizeof(password));
    echo_on();

    if(ret<0) {
      fprintf(stderr, "Error password\n");
      return 1;
    }

    password[ret]=0;

    fprintf(stderr, "\nOpen socket... ");
    if((ret=socket_daemon(sock, l))<1)
     return 1;

    fprintf(stderr, "done.\n");

    if(background)
      if(fork()) return 0;

    scheduler(ret, password, timer);

    socket_close(ret);

    unlink(sock);
  }

  /* Client */
  else {

    if((ret=socket_client(sock))<1)
      return 1;

    if(e)
       set_exit(ret);

    else {
      if(get_password(ret, password, sizeof(password))) {
        socket_close(ret);
        return 1;
      }

      if(mta_run(mta, password, stop, argc, argv)) {
        socket_close(ret);
        return 1;
      }
    }

    socket_close(ret);
  }

  return 0;
}

/* EOF */
